__id__ = "$Id$"

from trio import common
from trio.node import *

def execute(node, trio_cursor, isConfBase=True, isTemporary=False):
    name = node.children[0].value
    description = [[x.children[0].value, x.children[1].value] for x in node.children[1].children]
    u_columnnames = [x.value for x in node.children[2].children]
    type = node.children[3].value

    # Basic integrity check for uncertain attributes
    if not set(u_columnnames).issubset(set([x[0] for x in description])):
        raise Exception("Uncertainty set must be subset of columnlist")

    # Create data tables and full range of system indexes when there is no subquery
    create_trio_table(name, type, description, u_columnnames, trio_cursor, isConfBase, isTemporary)
    create_trio_indexes(name, type, trio_cursor, isTemporary)

    trio_cursor.isTemporary = isTemporary

def _get_columns(type, description, u_columnnames):
    v_cols = ", ".join([col[0] for col in description])
    c_cols = ", ".join(x[0] + " " + x[1] for x in description if not x[0] in u_columnnames)
    u_cols = ", ".join(x[0] + " " + x[1] for x in description if x[0] in u_columnnames)

    if type == common.TYPE_CERTAIN:
        v_cols += ", trio_xid, trio_xid as trio_aid"
        v_cols += ", 1 as trio_alts, 1 as trio_enc, 1.0 as trio_conf"
        c_cols += ", trio_xid int PRIMARY KEY"
    else:
        v_cols += ", c.trio_xid, trio_aid"
        v_cols += ", trio_alts, trio_enc, trio_conf"
        c_cols += (len(c_cols) > 0 and ", " or "")
        c_cols += "trio_xid int PRIMARY KEY, trio_alts int, trio_enc int"
        u_cols += (len(u_cols) > 0 and ", " or "")
        u_cols += "trio_xid int, trio_aid int PRIMARY KEY"
        u_cols += ", trio_conf double precision"

    return [v_cols, c_cols, u_cols]

def create_trio_table(name, type, description, u_columnnames, trio_cursor, isConfBase=False, isTemporary=False):
    tmpstr = isTemporary and "TEMPORARY" or ""
    i = 0
    while i < len(description):
        j = i + 1
        while j < len(description):
            if description[i][0] == description[j][0]:
                raise Exception("Duplicate or ambiguous column reference '%s'" % description[j][0])
            j += 1
        i += 1
    v_cols, c_cols, u_cols = _get_columns(type, description, u_columnnames)

    try:
        c = trio_cursor.tdb.pdb.cursor()

        # DDL commands
        if not isTemporary:
            c.execute("CREATE TABLE trio_c_%s (%s)" % (name, c_cols))
            if type == common.TYPE_CERTAIN:
                c.execute(("CREATE VIEW %s AS "
                           + "SELECT %s FROM trio_c_%s "
                           + "ORDER BY trio_xid"
                          ) % (name, v_cols, name))
            else:
                c.execute("CREATE TABLE trio_u_%s (%s)" % (name, u_cols))
                c.execute(("CREATE VIEW %s AS "
                           + "SELECT %s FROM trio_u_%s u, trio_c_%s c "
                           + "WHERE u.trio_xid = c.trio_xid "
                           + "ORDER BY c.trio_xid") % (name, v_cols, name, name))
        else:
            # CAUTION: temporary tables must use full range of Trio columns to work with 'DISTINCT'
            c.execute(("CREATE TEMPORARY TABLE %s "
                       + "(%s, trio_xid int, trio_aid int, trio_alts int, trio_enc int, trio_conf double precision"
                       + ")")
                      % (name, ", ".join(str(col[0]) + " " + str(col[1]) for col in description)))

        # Lineage table
        c.execute("CREATE %s TABLE trio_lin_%s (aid int, disjlin int, srclin int, srctable name)" % (tmpstr, name))
        if isConfBase and not isTemporary:
            c.execute("CREATE TABLE trio_x_lin_%s (aid int, attr name, value varchar(512))" % name)

        # Metadata for attributes/columns
        query = "INSERT INTO trio_attribs (tablename, attrname, colnum, iscertain) VALUES (%s, %s, %s, %s)"
        for i in range(len(description)):
            c.execute(query, (name, description[i][0], i, not description[i][0] in u_columnnames))

        # Metadata for tables
        query = "INSERT INTO trio_tables (tablename, type, confbase, istemporary) VALUES ('%s', %s, %s, %s)"
        c.execute(query % (name, type, isConfBase, isTemporary))

        # Identifier sequences
        c.execute("CREATE %s SEQUENCE trio_xid_%s" % (tmpstr, name))
        if type != common.TYPE_CERTAIN or isTemporary:
            c.execute("CREATE %s SEQUENCE trio_aid_%s" % (tmpstr, name))

        # Update rule for confidence memoization
        if type != common.TYPE_CERTAIN and not isTemporary:
            c.execute(("CREATE RULE trio_conf_%s AS ON UPDATE TO %s "
                       + "DO INSTEAD "
                       + "UPDATE trio_u_%s SET trio_conf = new.trio_conf "
                       + "WHERE trio_aid = old.trio_aid") % (name, name, name))

        c.close()

    except Exception, e:
        if 'already exists' in str(e):
            raise Exception("Table '%s' already exists in Trio schema!" % name)
        else:
            raise e

def create_trio_indexes(name, type, trio_cursor, isTemporary=False):
    c = trio_cursor.tdb.pdb.cursor()
    c.execute("CREATE INDEX trio_lin_%s_idx ON trio_lin_%s(aid, disjlin, srclin)" % (name, name))
    c.close()

def update_trio_sequences(name, trio_cursor):
    c = trio_cursor.tdb.pdb.cursor()
    c.execute("SELECT max(trio_xid)+1, max(trio_aid)+1 FROM %s" % name)
    [max_xid, max_aid] = c.fetchall()[0]
    if max_xid is None:
        max_xid = "1"
    if max_aid is None:
        max_aid = "1"

    # Identifier sequences
    c.execute("ALTER SEQUENCE trio_xid_%s RESTART %s" % (name, max_xid))
    if trio_cursor.tableType != common.TYPE_CERTAIN or trio_cursor.isTemporary:
        c.execute("ALTER SEQUENCE trio_aid_%s RESTART %s" % (name, max_aid))
    c.close()
