__id__ = "$Id$"

from trio import common
from trio import xtuple
import select
import createtable
from trio.node import *

class InsertExecutor:
    def __init__(self, table, trio_cursor, columnList=None, uncertainSet=None):
        self.table = table
        self.cursor = trio_cursor.tdb.pdb.cursor()
        self.select_list = ""

        # Some table metadata
        self.columnnames = []
        self.c_columnnames = []
        self.u_columnnames = []
        self.u_columns = []

        if columnList is not None or uncertainSet is None or trio_cursor.description is None:
            # Get the metadata from schema, table type first
            query = "SELECT type, confbase, istemporary FROM trio_tables WHERE tablename = '%s'"
            self.cursor.execute(query % self.table)
            tup = self.cursor.fetchone()
            if not tup:
                raise Exception("Table '%s' does not exist in Trio schema!" % self.table)
            trio_cursor.tableType = tup[0]
            trio_cursor.isConfbase = tup[1]
            trio_cursor.isTemporary = tup[2]

            # Column types
            query = "SELECT attrname, iscertain, colnum FROM trio_attribs WHERE tablename = '%s' ORDER BY colnum asc"
            self.cursor.execute(query % self.table)
            columns = self.cursor.fetchall()
            if not columns:
                raise Exception("Table '%s' does not exist in Trio schema!" % self.table)
            if columnList is None:
                self.columnnames = [x[0] for x in columns]
                self.c_columnnames = [x[0] for x in columns if x[1]]
                self.u_columnnames = [x[0] for x in columns if not x[1]]
                self.u_columns = [int(x[2]) for x in columns if not x[1]]
            else:
                i = 0
                for col in columnList:
                    matched = False
                    for x in columns:
                        if col == x[0]:
                            if col in self.columnnames:
                                raise Exception("Duplicate column '%s'" % col)
                            self.columnnames += [x[0]]
                            if x[1]:
                                self.c_columnnames += [x[0]]
                            else:
                                self.u_columnnames += [x[0]]
                                self.u_columns += [i]
                            i += 1
                            matched = True
                    if not matched:
                        raise Exception("Unresolvable column '%s'" % col)

            trio_cursor.u_columns = self.u_columns

        else:
            # Get the metadata directly from cursor
            i = 0
            for c in trio_cursor.description:
                if i == trio_cursor.num_datavalues:
                    break
                self.columnnames.append(c[0])
                if c[0] in uncertainSet:
                    self.u_columnnames.append(c[0])
                    self.u_columns.append(i)
                else:
                    self.c_columnnames.append(c[0])
                i += 1

        # Prepare insert statements
        if trio_cursor.isTemporary:
            self.insert_u_sql = (
                "INSERT INTO " + self.table + " ("
                + ", ".join(x for x in self.columnnames)
                + ", trio_xid, trio_aid, trio_alts, trio_enc, trio_conf"
                + ") VALUES ("
                + common.bindparamstr(len(self.columnnames)) + ", %d, %d, %d, %d, %s"
                + ")")
        else:
            if trio_cursor.tableType == common.TYPE_CERTAIN:
                self.insert_c_sql = (
                    "INSERT INTO trio_c_" + self.table + " ("
                    + ", ".join(x for x in self.c_columnnames)
                    + ", trio_xid"
                    + ") VALUES ("
                    + common.bindparamstr(len(self.c_columnnames)) + ", %d"
                    + ")")
            else:
                self.insert_c_sql = (
                    "INSERT INTO trio_c_" + self.table + " ("
                    + ", ".join(x for x in self.c_columnnames)
                    + (len(self.c_columnnames) > 0 and ", " or "")
                    + "trio_xid, trio_alts, trio_enc"
                    + ") VALUES ("
                    + common.bindparamstr(len(self.c_columnnames))
                    + (len(self.c_columnnames) > 0 and ", " or "") + "%d, %d, %d"
                    + ")")
                self.insert_u_sql = (
                    "INSERT INTO trio_u_" + self.table + " ("
                    + ", ".join(x for x in self.u_columnnames)
                    + (len(self.u_columnnames) > 0 and ", " or "")
                    + "trio_xid, trio_aid, trio_conf"
                    + ") VALUES ("
                    + common.bindparamstr(len(self.u_columnnames))
                    + (len(self.u_columnnames) > 0 and ", " or "") + "%d, %d, %s"
                    + ")")
        self.insert_lin_sql = "INSERT INTO trio_lin_%s (aid, disjlin, srclin, srctable) " % self.table \
                              + "VALUES (%d, %d, %d, %s)"
        self.insert_x_lin_sql = "INSERT INTO trio_x_lin_%s (aid, attr, value) " % self.table \
                                 + "VALUES (%d, %s, %s)"

    def execute(self, node, trio_cursor):
        # Insert statement with subquery
        if isinstance(node.children[2], Select):
            if not trio_cursor.setOperator:
                if not trio_cursor.isConfbase:
                    raise Exception("Inserts only allowed into base tables!")
                # Enforce NoLineage and ComputeConf modifiers
                if isinstance(node.children[2].children[0].children[0], Empty):
                    node.children[2].children[0].children[0] = Node('nolineage')
                else:
                    noLineage = False
                    for child in node.children[2].children[0].children:
                        if child.value == 'nolineage':
                            noLineage = True
                            break
                    if not noLineage:
                        node.children[2].children[0].children.append(Node('nolineage'))
                if trio_cursor.tableType == common.TYPE_CONFIDENCE \
                   and not isinstance(node.children[2].children[4], ComputeConfidences):
                    node.children[2].children[4] = ComputeConfidences()

            # Rewrite and run the SQL query
            sql = select.execute(node.children[2], trio_cursor, False)
            if len(trio_cursor.description) != len(self.columnnames):
                raise Exception("Columns do not match!")
            for x in trio_cursor.description:
                if x[0] not in self.columnnames:
                    raise Exception("Unresolvable column '%s'!" % x[0])

            # Check uncertain sets
            if not set(trio_cursor.u_columns).issubset(set(self.u_columns)):
                raise Exception("Uncertain attributes in subquery must be a subset of table's uncertain attributes!")
            trio_cursor.isFetchable = False

            if int(common.cp.get("global", "use_SPI")) == 0:
                # Python-Fetch
                while True:
                    xtup = trio_cursor.fetchone()
                    if xtup is None:
                        break
                    self.__checkTableType(xtup, trio_cursor)
                    self.insertXTuple(xtup, trio_cursor)

            else:
                # SPI-Fetch
                conf_option = 0
                if trio_cursor.noconf:
                    conf_option = -1
                if trio_cursor.gotAsConf:
                    conf_option = 1
                if trio_cursor.gotUniform:
                    conf_option = 2
                if not trio_cursor.distinct:
                    tables = "|".join("%s" % name for name in node.children[2].names) + "|"
                else:
                    tables = "%s|" % trio_cursor.distinctTable
                columns = "|".join(col for col in self.columnnames) + "|"
                coltypes = "|".join(i in self.u_columns and "u" or "c" for i in range(len(self.columnnames))) + "|"
                c = trio_cursor.tdb.pdb.cursor()
                c.execute(("SELECT * FROM trio_selectinto($$%s$$, $$%s$$, $$%s$$, $$%s$$, $$%s$$"
                           + ", %d, %d, %d, %d, %d, %d, %d, %d, %d)")
                          % (self.table, sql, tables, columns, coltypes, trio_cursor.tableType, trio_cursor.isTemporary,
                             trio_cursor.merged, trio_cursor.flatten, trio_cursor.groupAlts,
                             trio_cursor.nolineage, conf_option, trio_cursor.nomaybe, trio_cursor.distinct))
                if c.fetchone()[0] == 0:
                    raise Exception("SPI call failed!")
                c.close()

            createtable.update_trio_sequences(self.table, trio_cursor)

            # Re-use lineage from temp table for 'DISTINCT' in 'CREATE TABLE AS' statements
            if trio_cursor.distinct and not trio_cursor.nolineage:
                trio_cursor.cursor.execute("INSERT INTO trio_lin_%s (SELECT * FROM trio_lin_%s)"
                                           % (self.table, trio_cursor.distinctTable))

            # Update some lineage metadata
            if not trio_cursor.nolineage:
                lin_tables = node.children[2].names
                for lin_table in lin_tables:
                    try:
                        c.execute("DELETE FROM trio_tables_lin WHERE tablename = '%s' and srctable = '%s'"
                                  % (self.table, lin_table))
                    except Exception, e:
                        pass
                    trio_cursor.tdb.pdb.cursor().execute(("INSERT INTO trio_tables_lin (tablename, srctable) "
                                                          + "VALUES ('%s', '%s')") % (self.table, lin_table))

        # Manual insert statement
        else:
            trio_cursor.isFetchable = False
            if isinstance(node.children[1], SelectClause):
                self.select_list = [x.value for x in node.children[1].children]
            xtup = xtuple.XTuple()
            numAlts = len(node.children[2].children[0].children)
            questionMark = numAlts
            if isinstance(node.children[2].children[1], Question):
                questionMark = numAlts + 1
            for alt in node.children[2].children[0].children:
                xtup.append(xtuple.Alternative([x.value for x in alt.children],
                                               alts=numAlts, enc=questionMark, conf=alt.value))
            external_lineages = None
            if isinstance(node.children[3], ExternalLineageSpec):
                if len(xtup) != len(node.children[3].children):
                    raise Exception("The number of external lineages doesn't match the number of alternatives.")
                external_lineages = []
                for external_lineage in node.children[3].children:
                    external_lineages.append([(x.children[0].value, x.children[1].value)
                                              for x in external_lineage.children])
            self.__checkTableType(xtup, trio_cursor)
            self.insertXTuple(xtup, trio_cursor, external_lineages)

    def insertXTuple(self, xtup, trio_cursor, external_lineages=None):
        # Get a new X-ID
        if trio_cursor.tableType == common.TYPE_CERTAIN and trio_cursor.distinct and not trio_cursor.nolineage:
            newxid = xtup.alternatives[0].getLineage().items()[0][1]
        else:
            self.cursor.execute("SELECT nextval('trio_xid_%s')" % self.table)
            newxid = int(self.cursor.fetchone()[0])

        for i in range(len(xtup)):
            if trio_cursor.tableType == common.TYPE_CERTAIN:
                # Project X-ID onto A-ID
                newaid = newxid
            elif trio_cursor.distinct and not trio_cursor.nolineage:
                # Forward A-IDs from temp table (only for 'DISTINCT' in 'CREATE TABLE AS' statements)
                newaid = xtup.alternatives[i].getLineage().items()[0][1]
            else:
                # Get a new A-ID
                self.cursor.execute("SELECT nextval('trio_aid_%s')" % self.table)
                newaid = int(self.cursor.fetchone()[0])
            insertCertain = True
            for j in range(i):
                    insertCertain = False
                    break
            self.__insertAlternative(xtup.alternatives[i], newaid, newxid, trio_cursor, insertCertain,
                                     external_lineages is not None and external_lineages[i] or None)

    def __insertAlternative(self, alternative, aid, xid, trio_cursor, insertCertain=True, external_lineage=None):
        # Insert data wrt. encoding type
        u_cols = []
        c_cols = []
        if trio_cursor.isTemporary:
            for i in range(len(self.columnnames)):
                u_cols += [alternative[i]]
            u_cols += [xid, aid, alternative.alts, alternative.enc, alternative.confidence]
        else:
            for i in range(len(self.columnnames)):
                if i in self.u_columns:
                    u_cols += [alternative[i]]
                else:
                    c_cols += [alternative[i]]
            c_cols += [xid]
            u_cols += [xid]
            if trio_cursor.tableType != common.TYPE_CERTAIN:
                c_cols += [alternative.alts, alternative.enc]
                u_cols += [aid, alternative.confidence]

        # Execute statements
        if not trio_cursor.isTemporary and insertCertain:
            self.cursor.execute(self.insert_c_sql, c_cols)
        if trio_cursor.tableType != common.TYPE_CERTAIN or trio_cursor.isTemporary:
            self.cursor.execute(self.insert_u_sql, u_cols)

        # Insert lineage if any
        if not trio_cursor.distinct and alternative.getLineage():
            for lin in alternative.getLineage().items():
                self.cursor.execute(self.insert_lin_sql, (aid, lin[2], lin[1], lin[0]))

        # Insert external lineage if any
        if external_lineage is not None and not trio_cursor.isTemporary:
            for attr_value in external_lineage:
                self.cursor.execute(self.insert_x_lin_sql, (aid, attr_value[0], attr_value[1]))

    def __checkTableType(self, xtup, trio_cursor):
        if len(xtup) > 1 and not self.u_columns:
            raise Exception("Inserting alternatives not allowed into a table with no uncertain attributes!")
        if not trio_cursor.setOperator and not trio_cursor.isConfbase:
            raise Exception("Inserts only allowed into base tables")
        elif trio_cursor.tableType == common.TYPE_CONFIDENCE:
            if xtup.getConfidence() is None and trio_cursor.isConfbase:
                raise Exception("Confidence value is required for base table")
        elif trio_cursor.tableType == common.TYPE_UNCERTAIN:
            if xtup.getConfidence() is not None and xtup.getConfidence() < 1.0:
                raise Exception("Confidence value is forbidden for green table")
        elif trio_cursor.tableType == common.TYPE_CERTAIN:
            if len(xtup) > 1:
                raise Exception("Cannot insert alternatives into certain table type")
            if xtup.getConfidence() is not None and xtup.getConfidence() < 1.0:
                raise Exception("Confidence value is forbidden for blue table")
            if xtup.getQuestionMark():
                raise Exception("Question mark is forbidden for blue table")
        else:
            raise Exception("Unsupported type")
