__id__ = "$Id$"

from lexer import tokens
from ply import yacc
from trio.node import *
from trio.common import TYPE_CERTAIN, TYPE_UNCERTAIN, TYPE_CONFIDENCE
import tempfile
import sys

start = "statement"

#
# A statement
# Note: We do not handle EXCEPT as a select statement, since we do not allow more than one level of subqueries.
#
def p_statement(p):
    """statement : create_index_statement
                 | create_table_statement
                 | create_temp_table_statement
                 | create_table_as_statement
                 | create_temp_table_as_statement
                 | drop_index_statement
                 | drop_table_statement
                 | insert_statement
                 | insert_statement_subquery
                 | select_statement
                 | setop_statement
                 | view_table_statement
                 | update_statement
                 | delete_statement
                 | trio_command"""
    p[0] = p[1]

#
# create index statement
#
def p_create_index_statement(p):
    """create_index_statement : create index id on id '(' id_list ')'"""
    columns = IdentifierList(children=p[7])
    p[0] = CreateIndex(children=[Node(p[3]), Node(p[5]), columns])

#
# create (temporary) table statement
#
def p_create_table_statement(p):
    """create_table_statement : create trio table id '(' column_definition_list_uncertain ')' opt_withconf
                              | create table id '(' column_definition_list ')'"""
    if len(p) == 9:
        table = Node(p[4])
        type = Node()
        if isinstance(p[8], Empty):
            type.value = TYPE_UNCERTAIN
        else:
            type.value = TYPE_CONFIDENCE
        columns = ColumnList(children=p[6][0])
        p[0] = CreateTable(children=[table, columns, p[6][1], type])
    else:
        table = Node(p[3])
        type = Node(TYPE_CERTAIN)
        columns = ColumnList(children=p[5])
        p[0] = CreateTable(children=[table, columns, UncertainSet(), type])

def p_create_temp_table_statement(p):
    """create_temp_table_statement : create temporary trio table id '(' column_definition_list_uncertain ')' opt_withconf
                                   | create temporary table id '(' column_definition_list ')'"""
    if len(p) == 10:
        table = Node(p[5])
        type = Node()
        if isinstance(p[9], Empty):
            type.value = TYPE_UNCERTAIN
        else:
            type.value = TYPE_CONFIDENCE
        columns = ColumnList(children=p[7][0])
        p[0] = CreateTempTable(children=[table, columns, p[7][1], type])
    else:
        table = Node(p[4])
        type = Node(TYPE_CERTAIN)
        columns = ColumnList(children=p[6])
        p[0] = CreateTempTable(children=[table, columns, UncertainSet(), type])

def p_column_definition_list_uncertain(p):
    """column_definition_list_uncertain : column_definition_list ',' uncertain '(' column_name_list ')'
                                        | column_definition_list"""
    if len(p) == 2:
        p[0] = [p[1], UncertainSet()]
    else:
        p[0] = [p[1], UncertainSet(children=p[5])]

def p_column_definition_list(p):
    """column_definition_list : column_definition_list ',' column_definition
                              | column_definition"""
    if len(p) == 4:
        p[0] = p[1] + [p[3]]
    else:
        p[0] = [p[1]]

def p_column_definition(p):
    """column_definition : id data_type"""
    p[0] = Column(children=[Node(p[1]), p[2]])

def p_data_type(p):
    """data_type : int
                 | int4
                 | int8
                 | float
                 | float4
                 | float8
                 | varchar opt_length
                 | text"""
    if len(p) == 2:
        p[0] = DataType(p[1])
    else:
        p[0] = DataType(p[1], [p[2]])

def p_opt_length(p):
    """opt_length : '(' numeric_literal ')'
                  | empty"""
    if len(p) == 4:
        p[0] = Node(p[2])
    else:
        p[0] = p[1]

def p_opt_withconf(p):
    """opt_withconf : with confidences
                    | empty"""
    if len(p) == 3:
        p[0] = WithConfidences()
    else:
        p[0] = p[1]

#
# create (temporary) table as statement
#
def p_create_table_as_statement(p):
    """create_table_as_statement : create table id as select_statement
                                 | create table id as setop_statement"""
    p[0] = CreateTableAs(children=[Node(p[3]), p[5]])

def p_create_temp_table_as_statement(p):
    """create_temp_table_as_statement : create temporary table id as select_statement
                                      | create temporary table id as setop_statement"""
    p[0] = CreateTempTableAs(children=[Node(p[4]), p[6]])

#
# drop index statement
#
def p_drop_index_statement(p):
    """drop_index_statement : drop index id"""
    p[0] = DropIndex(children=[Node(p[3])])

#
# drop table statement
#
def p_drop_table_statement(p):
    """drop_table_statement : drop table id_list opt_cascade"""
    tables = DropTableList(children=p[3])
    p[0] = DropTable(children=[tables, p[4]])

def p_opt_cascade(p):
    """opt_cascade : cascade
                   | empty"""
    if p[1] == "cascade":
        p[0] = Cascade(p[1])
    else:
        p[0] = p[1]

#
# insert statement
#
def p_insert_statement(p):
    """insert_statement : insert into id values insert_spec opt_extlineage
                        | insert into id '(' column_name_list ')' values insert_spec opt_extlineage"""
    if len(p) == 10:
        p[0] = Insert(children=[Node(p[3]), ColumnList(children=p[5]), p[8], p[9]])
    else:
        p[0] = Insert(children=[Node(p[3]), Empty(), p[5], p[6]])

def p_insert_statement_subquery(p):
    """insert_statement_subquery : insert into id select_statement
                                 | insert into id '(' select_list ')' select_statement"""
    if len(p) == 5:
        p[0] = Insert(children=[Node(p[3]), Empty(), p[4]])
    else:
        p[0] = Insert(children=[Node(p[3]), SelectClause(children=p[5]), p[7]])

def p_insert_spec(p):
    """insert_spec : '[' alt_altlist ']' opt_question
                   | '(' insert_value_list ')' opt_question
                   | '[' alt_conf_altlist ']'
                   | '(' insert_value_conf_list ')'"""
    if isinstance(p[2], Select):
        p[0] = InsertSpec(children=[p[2]])
    else:
        alts = InsertList(children=p[2])
        if len(p) == 5:
            p[0] = InsertSpec(children=[alts, p[4]])
        else:
            conf = sum([x.value for x in alts.children])
            if conf < 1.0:
                p[0] = InsertSpec(children=[alts, Question()])
            else:
                p[0] = InsertSpec(children=[alts, Empty()])

def p_alt_altlist(p):
    """alt_altlist : alt_altlist '|' '(' literal_list ')'
                   | '(' literal_list ')'"""
    if len(p) == 6:
        p[0] = p[1] + [Alternative(None, p[4])]
    else:
        p[0] = [Alternative(None, p[2])]

def p_insert_value_list(p):
    """insert_value_list : insert_value_list ',' insert_value
                         | insert_value"""
    if len(p) == 4:
        p[0] = [Alternative(None, x.children + y.children) for x in p[1] for y in p[3]]
    else:
        p[0] = p[1]

def p_insert_value(p):
    """insert_value : literal
                    | '[' literal_altlist ']'"""
    if len(p) == 2:
        p[0] = [Alternative(None, [p[1]])]
    else:
        p[0] = [Alternative(None, [x]) for x in p[2]]

def p_literal_altlist(p):
    """literal_altlist : literal_altlist '|' literal
                       | literal"""
    if len(p) == 4:
        p[0] = p[1] + [p[3]]
    else:
        p[0] = [p[1]]

def p_alt_conf_altlist(p):
    """alt_conf_altlist : alt_conf_altlist '|' '(' literal_list ')' ':' numeric_literal
                        | '(' literal_list ')' ':' numeric_literal"""
    if len(p) == 8:
        p[0] = p[1] + [Alternative(p[7], children=p[4])]
    else:
        p[0] = [Alternative(p[5], children=p[2])]

def p_insert_value_conf_list(p):
    """insert_value_conf_list : insert_value_conf_list ',' insert_value_conf
                              | insert_value_conf"""
    if len(p) == 4:
        p[0] = [Alternative(x.value * y.value, x.children + y.children) for x in p[1] for y in p[3]]
    else:
        p[0] = p[1]

def p_insert_value_conf(p):
    """insert_value_conf : literal_conf
                         | '[' literal_conf_altlist ']'"""
    if len(p) == 4:
        p[0] = [Alternative(x[1], [x[0]]) for x in p[2]]
    else:
        p[0] = [Alternative(p[1][1], [p[1][0]])]

def p_literal_conf(p):
    """literal_conf : literal ':' numeric_literal"""
    p[0] = [p[1], p[3]]

def p_literal_conf_altlist(p):
    """literal_conf_altlist : literal_conf_altlist '|' literal_conf
                            | literal_conf"""
    if len(p) == 4:
        p[0] = p[1] + [p[3]]
    else:
        p[0] = [p[1]]

def p_opt_question(p):
    """opt_question : '?'
                    | empty"""
    if p[1] == '?':
        p[0] = Question()
    else:
        p[0] = p[1]

def p_opt_extlineage(p):
    """opt_extlineage : extlineage extlineage_spec
                      | empty"""
    if len(p) == 2:
        p[0] = p[1]
    else:
        p[0] = p[2]

def p_extlineage_spec(p):
    """extlineage_spec : '[' attr_value_set_altlist ']'
                       | '(' attr_value_list ')'"""
    if p[1] == '[':
        p[0] = ExternalLineageSpec(children=p[2])
    else:
        p[0] = ExternalLineageSpec(children=[Alternative(None, p[2])])

def p_attr_value_set_altlist(p):
    """attr_value_set_altlist : attr_value_set_altlist '|' '(' attr_value_list ')'
                              | attr_value_set_altlist '|' '(' empty ')'
                              | '(' attr_value_list ')'
                              | '(' empty ')'"""
    if len(p) == 6:
        if isinstance(p[4], Empty):
            p[0] = p[1] + [Alternative(None, [])]
        else:
            p[0] = p[1] + [Alternative(None, p[4])]
    else:
        if isinstance(p[2], Empty):
            p[0] = [Alternative(None, [])]
        else:
            p[0] = [Alternative(None, p[2])]

def p_attr_value_list(p):
    """attr_value_list : attr_value_list ',' attr_value
                       | attr_value"""
    if len(p) == 4:
        p[0] = p[1] + [p[3]]
    else:
        p[0] = [p[1]]

def p_attr_value(p):
    """attr_value : id ':' string_literal
                  | id ':' null"""
    p[0] = ExternalLineage(children=[Node(p[1]), Literal(p[3])])

#
# select statement
#
def p_select_statement(p):
    """select_statement : '(' select_statement ')'
                        | selection"""
    if len(p) == 4:
        p[0] = p[2]
    else:
        p[0] = p[1]

def p_selection(p):
    """selection : select select_option_list select_clause from table_list opt_where opt_version opt_groupbykeys opt_orderby opt_compute_conf"""
    so = SelectOptions(children=p[2])
    #sc = SelectClause(children=p[3])
    sc = p[3]
    fc = FromClause(children=p[5])
    p[0] = Select(children=[so, sc, fc, p[6], p[10], p[8], p[9], Empty(), p[7]])
    p[0].isViewTable = False

# select_option_list
def p_select_option_list(p):
    """select_option_list : select_option_list select_option
                          | select_option"""
    if len(p) == 3:
        p[0] = p[1] + [p[2]]
    else:
        p[0] = [p[1]]

def p_select_option(p):
    """select_option : distinct
                     | merged
                     | flatten
                     | groupalts '(' groupalts_list ')'
                     | nolineage
                     | noconf
                     | nomaybe
                     | empty"""
    if p[1] == "groupalts":
        p[0] = GroupAlts(children=[SelectClause(children=p[3])])
    else:
        if not isinstance(p[1], Empty):
            p[0] = Node(p[1])
        else:
            p[0] = Empty()

def p_groupalts_list(p):
    """groupalts_list : groupalts_asterisk
                      | groupalts_columnlist"""
    p[0] = p[1]

def p_groupalts_asterisk(p):
    """groupalts_asterisk : asterisk"""
    p[0] = [p[1]]

def p_groupalts_columnlist(p):
    """groupalts_columnlist : column_name_list"""
    p[0] = p[1]

# select_clause
def p_select_clause(p):
    """select_clause : select_list
                     | select_list '|' select_list"""
    if len(p) == 4:
        p[0] = ConditionalSelectClause(value=len(p[1]), children=p[1] + p[3])
    else:
        p[0] = SelectClause(children=p[1])

def p_select_list(p):
    """select_list : select_list ',' select_attribute
                   | select_attribute"""
    if len(p) == 4:
        p[0] = p[1] + [p[3]]
    else:
        p[0] = [p[1]]

def p_select_attribute(p):
    """select_attribute : asterisk
                        | expression
                        | expression as id
                        | expression as conf
                        | uniform as conf
                        | scaled_conf as conf
                        | horizontal_select
                        | horizontal_select as id
                        | horizontal_select as conf"""
    if len(p) == 2:
        p[0] = p[1]
    elif p[1] == "uniform":
        p[0] = As(children=[Node(p[1]), Node(p[3])])
    else:
        p[0] = As(children=[p[1], Node(p[3])])

def p_asterisk(p):
    """asterisk : '*'"""
    p[0] = Star()

def p_scaled(p):
    """scaled_conf : scaled
                   | scaledbyexp '(' expression ')'"""
    if len(p) == 2:
        p[0] = ConfScaled(children=[Empty()])
    elif len(p) == 5:
        p[0] = ConfScaled(children=[p[3]])

def p_expression(p):
    """expression : term
                  | expression '+' term
                  | expression '-' term"""
    if len(p) == 2:
        p[0] = p[1]
    else:
        p[0] = BinaryExpression(p[2], [p[1], p[3]])

def p_term(p):
    """term : factor
            | term '*' factor
            | term '/' factor"""
    if len(p) == 2:
        p[0] = p[1]
    else:
        p[0] = BinaryExpression(p[2], [p[1], p[3]])

def p_factor(p):
    """factor : primary
              | '+' factor
              | '-' factor"""
    if len(p) == 2:
        p[0] = p[1]
    else:
        p[0] = UnaryExpression(p[1], [p[2]])

def p_primary(p):
    """primary : column_name
               | literal
               | '(' expression ')'
               | horizontal_select
               | external_function_call
               | function_call"""
    if len(p) == 2:
        p[0] = p[1]
    else:
        p[0] = p[2]

def p_external_function_call(p):
    """external_function_call : id '(' empty ')'
                              | id '(' function_arguments ')'"""
    if isinstance(p[3], Empty):
        p[0] = p[0] = Function(p[1],[])
    else:
        p[0] = Function(p[1],p[3])

def p_function_arguments(p):
    """function_arguments : literal_list
                          | column_name_list
                          | id_list
                          | literal_list ',' function_arguments
                          | column_name_list ',' function_arguments
                          | id_list ',' function_arguments"""
    if len(p) == 2:
        p[0] = p[1]
    else:
        p[0] = p[1] + p[3]

def p_function_call(p):
    """function_call : lmin '(' expression ')'
                     | lmax '(' expression ')'
                     | lsum '(' expression ')'
                     | lavg '(' expression ')'
                     | lcount '(' expression ')'
                     | lcount '(' '*' ')'
                     | havg '(' expression ')'
                     | hcount '(' expression ')'
                     | hcount '(' '*' ')'
                     | hmax '(' expression ')'
                     | hmin '(' expression ')'
                     | hsum '(' expression ')'
                     | esum '(' expression ')'
                     | ecount '(' expression ')'
                     | ecount '(' '*' ')'
                     | emin '(' expression ')'
                     | emax '(' expression ')'
                     | eavg '(' expression ')'
                     | sum '(' expression ')'
                     | count '(' expression ')'
                     | count '(' asterisk ')'
                     | count '(' distinct expression ')'
                     | min '(' expression ')'
                     | max '(' expression ')'
                     | avg '(' expression ')'
                     | extlineage_function_call
                     | conf_function_call"""
    if len(p) == 5 or len(p) == 6:
        if p[1] == "lmin":
            p[0] = Aggregate(children=[Lmin(children=[p[3]])])
        elif p[1] == "lmax":
            p[0] = Aggregate(children=[Lmax(children=[p[3]])])
        elif p[1] == "lsum":
            p[0] = Aggregate(children=[Lsum(children=[p[3]])])
        elif p[1] == "lavg":
            p[0] = Aggregate(children=[Lavg(children=[p[3]])])
        elif p[1] == "lcount":
            if p[3] == '*':
                p[0] = Aggregate(children=[LcountStar()])
            else:
                raise Exception("Unsupported lcount aggregate function; only support lcount(*)")
        elif p[1] == "havg":
            p[0] = Aggregate(children=[Havg(children=[p[3]])])
        elif p[1] == "hcount":
            if p[3] == '*':
                p[0] = Aggregate(children=[HcountStar()])
            else:
                raise Exception("Unsupported hcount aggregate function; only support hcount(*)")
        elif p[1] == "hmax":
            p[0] = Aggregate(children=[Hmax(children=[p[3]])])
        elif p[1] == "hmin":
            p[0] = Aggregate(children=[Hmin(children=[p[3]])])
        elif p[1] == "hsum":
            p[0] = Aggregate(children=[Hsum(children=[p[3]])])
        elif p[1] == "esum":
            p[0] = Aggregate(children=[Esum(children=[p[3]])])
        elif p[1] == "ecount":
            if p[3] == '*':
                p[0] = Aggregate(children=[EcountStar()])
            else:
                p[0] = Aggregate(children=[Ecount(children=[p[3]])])
        elif p[1] == "emin":
            p[0] = Aggregate(children=[Emin(children=[p[3]])])
        elif p[1] == "emax":
            p[0] = Aggregate(children=[Emax(children=[p[3]])])
        elif p[1] == "eavg":
            p[0] = Aggregate(children=[Eavg(children=[p[3]])])
        elif p[1] == "max":
            p[0] = Aggregate(children=[Max(children=[p[3]])])
        elif p[1] == "min":
            p[0] = Aggregate(children=[Min(children=[p[3]])])
        elif p[1] == "count":
            if len(p) == 5:
                p[0] = Aggregate(children=[Count(children=[p[3]])])
            else:
                p[0] = Aggregate(children=[Count(children=[Node(p[3]), p[4]])])
        elif p[1] == "sum":
            p[0] = Aggregate(children=[Sum(children=[p[3]])])
        elif p[1] == "avg":
            p[0] = Aggregate(children=[Avg(children=[p[3]])])
        else:
            raise Exception("Unsupported aggregate function: " + p[1])
    else:
        p[0] = p[1]

def p_extlineage_function_call(p):
    """extlineage_function_call : extlineage '(' id ',' id ')'"""
    p[0] = ExternalLineage(children=[Node(p[3]), Node(p[5])])

def p_conf_function_call(p):
    """conf_function_call : conf '(' id_list ')'
                          | conf '(' asterisk ')'"""
    if isinstance(p[3], list):
        p[0] = Conf(children=p[3])
    else:
        p[0] = Conf(children=[Star()])

def p_horizontal_select(p):
    """horizontal_select : '[' select select_list from table_list opt_where opt_groupby opt_having ']'
                         | '[' select_list opt_where ']'
                         | '[' search_condition ']'"""
    if len(p) == 10:
        sc = SelectClause(children=p[3])
        fc = FromClause(children=p[5])
        p[0] = HorizontalSelect(children=[sc, fc, p[6], p[7], p[8]])
    elif len(p) == 5:
        sc = SelectClause(children=p[2])
        p[0] = HorizontalSelect(children=[sc, Empty(), p[3], Empty(), Empty()])
    else:
        sc = SelectClause(children=[Star()])
        wc = WhereClause(children=[p[2]])
        p[0] = HorizontalSelect(children=[sc, Empty(), wc, Empty(), Empty()])

def p_opt_groupby(p):
    """opt_groupby : group by column_name_list
                   | empty"""
    if len(p) == 4:
        p[0] = GroupbyClause(children=p[3])
    else:
        p[0] = p[1]

def p_opt_having(p):
    """opt_having : having search_condition
                  | empty"""
    if len(p) == 3:
        p[0] = HavingClause(children=[p[2]])
    else:
        p[0] = p[1]

# table_list
def p_table_list(p):
    """table_list : table_list ',' table_name
                  | table_name"""
    if len(p) == 4:
        p[0] = p[1] + [p[3]]
    else:
        p[0] = [p[1]]

def p_table_name(p):
    """table_name : base_table_name
                  | base_table_name id
                  | base_table_name as id"""
    if len(p) == 2:
        p[0] = p[1]
    elif len(p) == 3:
        p[0] = As(children=[p[1], Node(p[2])])
    else:
        p[0] = As(children=[p[1], Node(p[3])])

def p_base_table_name(p):
    """base_table_name : id
                       | uniform id"""
    if len(p) == 2:
        p[0] = Node(p[1])
    else:
        p[0] = UniformTable(p[2])

# opt_where
def p_opt_where(p):
    """opt_where : where search_condition
                 | empty"""
    if len(p) == 3:
        p[0] = WhereClause(children=[p[2]])
    else:
        p[0] = p[1]

def p_search_condition(p):
    """search_condition : boolean_term
                        | search_condition or boolean_term"""
    if len(p) == 2:
        p[0] = p[1]
    else:
        p[0] = BinaryExpression(p[2], [p[1], p[3]])

def p_boolean_term(p):
    """boolean_term : boolean_factor
                    | boolean_term and boolean_factor"""
    if len(p) == 2:
        p[0] = p[1]
    else:
        p[0] = BinaryExpression(p[2], [p[1], p[3]])

def p_boolean_factor(p):
    """boolean_factor : boolean_primary
                      | not boolean_primary"""
    if len(p) == 2:
        p[0] = p[1]
    else:
        p[0] = UnaryExpression(p[1], [p[2]])

def p_boolean_primary(p):
    """boolean_primary : predicate
                       | '(' search_condition ')'"""
    if len(p) == 2:
        p[0] = p[1]
    else:
        p[0] = Parenthesis(children=[p[2]])

def p_predicate(p):
    """predicate : comparison_predicate
                 | lineage_predicate
                 | extraneous_predicate
                 | exists_predicate
                 | not_exists_predicate
                 | in_predicate
                 | not_in_predicate
                 | maybe_predicate"""
    p[0] = p[1]

def p_comparison_predicate(p):
    """comparison_predicate : expression comparison expression
                            | expression like expression
                            | expression is expression
                            | expression is not expression
                            | expression comparison horizontal_select
                            | horizontal_select comparison expression
                            | expression comparison set_comparison_modifier horizontal_select"""
    if len(p) == 4:
        p[0] = BinaryExpression(p[2], [p[1], p[3]])
    else:
        p[0] = BinaryExpression(p[2] + " " + p[3], [p[1], p[4]])

def p_comparison(p):
    """comparison : '='
                  | comparison_op"""
    p[0] = p[1]

def p_set_comparison_modifier(p):
    """set_comparison_modifier : all
                               | any"""
    p[0] = p[1]

def p_lineage_predicate(p):
    """lineage_predicate : lineage '(' id ',' id ')'
                         | id lineage_arrow id"""
    if len(p) == 7:
        p[0] = Lineage(children=[Node(p[3]), Node(p[5])])
    else:
        p[0] = Lineage(children=[Node(p[1]), Node(p[3])])

def p_extraneous_predicate(p):
    """extraneous_predicate : extraneous '(' id_list ')'
                            | extraneous '(' asterisk ')'"""
    if isinstance(p[3], list):
        p[0] = Extraneous(children=p[3])
    else:
        p[0] = Extraneous(children=[Star()])

def p_exists_predicate(p):
    """exists_predicate : exists horizontal_select"""
    p[0] = UnaryExpression(p[1], [p[2]])

def p_not_exists_predicate(p):
    """not_exists_predicate : expression not exists horizontal_select"""
    p[0] = BinaryExpression(p[2] + " " + p[3], [p[1], p[4]])

def p_in_predicate(p):
    """in_predicate : expression in horizontal_select"""
    p[0] = BinaryExpression(p[2], [p[1], p[3]])

def p_not_in_predicate(p):
    """not_in_predicate : expression not in horizontal_select"""
    p[0] = BinaryExpression(p[2] + " " + p[3], [p[1], p[4]])

def p_maybe_predicate(p):
    """maybe_predicate : maybe '(' id ')'"""
    p[0] = Maybe(value=p[3])

# opt_version
def p_opt_version(p):
    """opt_version : at version numeric_literal
                   | at current version
                   | at any version
                   | empty"""
    if len(p) == 4:
        if p[2] == 'current' or p[2] == 'any':
            p[0] = Version(p[2])
        else:
            p[0] = Version(p[3])
    else:
        p[0] = Version("")

# opt_groupbykeys
def p_opt_groupbykeys(p):
    """opt_groupbykeys : group by column_name_list
                       | empty"""
    if len(p) == 4:
        p[0] = GroupByKey(children=p[3])
    else:
        p[0] = p[1]

# opt_orderby
def p_opt_orderby(p):
    """opt_orderby : order by orderby_list
                   | empty"""
    if len(p) == 4:
        p[0] = OrderBy(value="",children=p[3])
    else:
        p[0] = p[1]

def p_orderby_list(p):
    """orderby_list : orderby_list ',' column_name
                    | orderby_list ',' column_name asc
                    | orderby_list ',' column_name desc
                    | orderby_list ',' confidences
                    | orderby_list ',' confidences asc
                    | orderby_list ',' confidences desc
                    | column_name
                    | column_name asc
                    | column_name desc
                    | confidences
                    | confidences asc
                    | confidences desc"""
    if len(p) == 4:
        if isinstance(p[3], str):
            p[0] = p[1] + [OrderById(value="",children=[Conf(children=[Star()])])]
        else:
            p[0] = p[1] + [OrderById(value="",children=[p[3]])]
    elif len(p) == 5:
        if isinstance(p[3], str):
            p[0] = p[1] + [OrderById(value=p[4],children=[Conf(children=[Star()])])]
        else:
            p[0] = p[1] + [OrderById(value=p[4],children=[p[3]])]
    elif len(p) == 2:
        if isinstance(p[1], str):
            p[0] = [OrderById(value="",children=[Conf(children=[Star()])])]
        else:
            p[0] = [OrderById(value="",children=[p[1]])]
    else:
        if isinstance(p[1], str):
            p[0] = [OrderById(value=p[2],children=[Conf(children=[Star()])])]
        else:
            p[0] = [OrderById(value=p[2],children=[p[1]])]

# opt_compute_conf
def p_opt_compute_conf(p):
    """opt_compute_conf : compute confidences
                        | empty"""
    if len(p) == 3:
        p[0] = ComputeConfidences()
    else:
        p[0] = p[1]

#
# setop statement
#
def p_setop_statement(p):
    """setop_statement : '(' select_statement ')' except '(' select_statement ')' opt_compute_conf
                       | '(' select_statement ')' intersect '(' select_statement ')' opt_compute_conf
                       | '(' select_statement ')' union '(' select_statement ')' opt_compute_conf
                       | '(' select_statement ')' union all '(' select_statement ')' opt_compute_conf"""
    if len(p) == 9:
        p[0] = SetOperator(str(p[4]), children=[p[2], p[6], p[8]])
    else:
        p[0] = SetOperator(str(p[4]) + " " + str(p[5]), children=[p[2], p[7], p[9]])

#
# view table statement
#
def p_view_table_statement(p):
    """view_table_statement : view table id opt_viewtable_where opt_version opt_compute_conf opt_explain_lineage"""
    so = SelectOptions(children=[])
    sc = SelectClause(children=[Star()])
    fc = FromClause(children=[Node(p[3])])
    p[0] = Select(children=[so, sc, fc, p[4], p[6], Empty(), Empty(), Empty(), p[5]])
    p[0].isViewTable = True
    p[0].hasExplainLineage = isinstance(p[7], ExplainLineage)

def p_opt_viewtable_where(p):
    """opt_viewtable_where : where search_condition
                           | where trio_aid comparison literal
                           | where trio_xid comparison literal
                           | where trio_xid comparison literal and trio_aid comparison literal
                           | where trio_aid comparison literal and trio_xid comparison literal
                           | empty"""
    if len(p) == 3:
        p[0] = WhereClause(children=[p[2]])
    elif len(p) == 5:
        p[0] = WhereClause(children=[BinaryExpression(p[3], [Node(p[2]), p[4]])])
    elif len(p) == 9:
        p[0] = WhereClause(children=[BinaryExpression('and', [\
                           BinaryExpression(p[3], [Node(p[2]), p[4]]),\
                           BinaryExpression(p[7], [Node(p[6]), p[8]])])])
    else:
        p[0] = p[1]

def p_opt_explain_lineage(p):
    """opt_explain_lineage : explain lineage
                           | empty"""
    if len(p) == 3:
        p[0] = ExplainLineage()
    else:
        p[0] = p[1]

#
# trio command
#
def p_trio_command(p):
    """trio_command : command opt_args"""
    p[0] = Command(p[1], p[2])

def p_opt_args(p):
    """opt_args : opt_args  id
                | id
                | empty"""
    if len(p) == 3:
        p[0] = p[1] + [Node(p[2])]
    elif not isinstance(p[1], Empty):
        p[0] = [Node(p[1])]
    else:
        p[0] = [p[1]]

#
# update statement
#
def p_update_statement(p):
    """update_statement : update id set update_set_list opt_maybe opt_where
                        | update id addmaybe opt_where
                        | update id delmaybe opt_where"""
    if len(p) == 7:
        p[0] = Update(children=[Node(p[2]), UpdateSetClause(children=p[4]), p[5], p[6]])
    else:
        p[0] = Update(children=[Node(p[2]), Empty(), Node(p[3]), p[4]])

def p_update_set_list(p):
    """update_set_list : update_set_list ',' set_assignment
                       | set_assignment"""
    if len(p) == 4:
        p[0] = p[1]+[p[3]]
    else:
        p[0] = [p[1]]

def p_set_assignment(p):
    """set_assignment : column_name '=' expression
                      | conf '=' expression
                      | lineage '=' expression"""
    if p[1] == 'lineage':
        p[0] = LineageAssignment(children=[p[3]])
    elif p[1] == 'conf':
        p[0] = ConfAssignment(children=[p[3]])
    else:
        p[0] = ValueAssignment(children=[p[1],p[3]])

def p_opt_maybe(p):
    """opt_maybe : addmaybe
                 | delmaybe
                 | empty"""
    if isinstance(p[1], Empty):
        p[0] = p[1]
    else:
        p[0] = Node(p[1])

#
# delete statement
#
def p_delete_statement(p):
    """delete_statement : delete from id opt_maybe opt_where opt_cascade"""
    p[0] = Delete(children=[Node(p[3]), Empty(), p[4], p[5], p[6]])

#
# common
#
def p_empty(p):
    """empty :"""
    p[0] = Empty("")

def p_id_list(p):
    """id_list : id_list ',' id
               | id"""
    if len(p) == 4:
        p[0] = p[1] + [Node(p[3])]
    else:
        p[0] = [Node(p[1])]

def p_literal(p):
    """literal : numeric_literal
               | string_literal
               | null"""
    p[0] = Literal(p[1])

def p_literal_list(p):
    """literal_list : literal_list ',' literal
                    | literal"""
    if len(p) == 4:
        p[0] = p[1] + [p[3]]
    else:
        p[0] = [p[1]]

def p_column_name(p):
    """column_name : id
                   | id '.' id"""
    if len(p) == 2:
        p[0] = Node(p[1])
    else:
        p[0] = Dot(value=str(p[1]) + "." + str(p[2]), children=[Node(p[1]), Node(p[3])])

def p_column_name_list(p):
    """column_name_list : column_name_list ',' column_name
                        | column_name"""
    if len(p) == 4:
        p[0] = p[1] + [p[3]]
    else:
        p[0] = [p[1]]

#### Catastrophic error handler
def p_error(p):
    if not p:
        raise Exception("Syntax error")
    else:
        raise Exception("Syntax error near '%s'" % p.value)

parsetab_dir = tempfile.gettempdir()
sys.path.append(parsetab_dir)
parser = yacc.yacc(debug=0, tabmodule="trioparsetab", outputdir=parsetab_dir)

def parse(data):
    return parser.parse(data)
