__id__ = "$Id$"

import pgdb
from trio import common
from trio import xtuple

from trio.node import *
from trio.visitor import dumper
from trio.visitor import printer
from trio.executor import executor
from trio.visitor import restructurer
from trio.parser import parser

def connect(**options):
    pdb = pgdb.connect(**options)
    try:
        pdb.cursor().execute("DELETE FROM trio_tables_lin WHERE tablename IN "
                             + "(SELECT DISTINCT tablename FROM trio_tables WHERE istemporary = true)")
        pdb.cursor().execute("DELETE FROM trio_attribs WHERE tablename IN "
                             + "(SELECT DISTINCT tablename FROM trio_tables WHERE istemporary = true)")
        pdb.cursor().execute("DELETE FROM trio_tables WHERE istemporary = true")
        pdb.commit()
    except Exception, e:
        print "\nWARNING: Trio metadata inconsistent or not initialized!"
        pdb.rollback()
    return TrioCnx(pdb, options)

class TrioCnx:
    def __init__(self, pdb, options):
        self.pdb = pdb
        self.options = options

    def ncursor(self):
        return self.pdb.cursor()

    def cursor(self):
        return TrioCursor(self)

    def commit(self):
        self.pdb.commit()

    def rollback(self):
        self.pdb.rollback()

    def close(self):
        self.pdb.close()

class TrioCursor:
    def __init__(self, tdb):
        self.tdb = tdb
        self.num_joins = 0
        self.num_datavalues = 0
        self.root_node = None
        self.cursor = None
        self.sqlCommand = None
        self.setOperator = False
        self.debuglevel = int(common.cp.get("global", "debug_level"))

    def execute(self, sqlCommand):
        self.isXCursor = False
        self.isFetchable = False
        self.lastAlt = None
        self.lastXid = None
        self.isTemporary = False

        self.gotUniform = False
        self.isConfBase = False
        self.gotAsConf = False
        self.merged = False
        self.flatten = False
        self.noconf = False
        self.nomaybe = False
        self.nolineage = False
        self.groupAlts = -1
        self.distinct = False
        self.conditionalIdx = 0

        self.sqlCommand = sqlCommand
        s = parser.parse(sqlCommand)

        # Initial TriQL parse tree
        if int(common.cp.get("global", "debug_level")) > 0:
            print self.sqlCommand
            print
            dumper.Dumper().visit(s)
            print

        r = restructurer.Restructurer()
        t = r.visit(s)

        # Rewritten TriQL parse tree
        if int(common.cp.get("global", "debug_level")) > 1:
            print
            dumper.Dumper().visit(t)
            print

        e = executor.Executor(self)
        e.execute(s)

    def __getattr__(self, attrname):
        if self.cursor is None:
            return None
        if attrname == "rowcount":
            return self.cursor.rowcount
        elif attrname == "description":
            return self.cursor.description
        elif attrname == "tabletype":
            return self.cursor.tabletype

    def fetchall(self):
        return self.cursor.fetchall()

    def fetchone(self):
        if self.isXCursor:
            xtup = self.xfetchone()
            return xtup
        else:
            return self.cursor.fetchone()

    def xfetchall(self):
        tuples = []
        while True:
            tuple = self.xfetchone()
            if tuple is None:
                break
            tuples.append(tuple)
        return tuples

    def xfetchone(self):
        # Create a new xtuple
        xtup = xtuple.XTuple()

        # Fetch last alternative from buffer
        if self.lastAlt is not None:
            xtup.xid = self.lastXid
            xtup.append(self.__parseAlternative(self.lastAlt))

        # Keep on fetching more alternatives
        while True:
            self.lastAlt = self.cursor.fetchone()
            if self.lastAlt is None:
                if len(xtup) == 0:
                    return None
                else:
                    break

            if self.groupAlts == -1:
                xid = self.__getXid(self.lastAlt)
            else:
                xid = self.lastAlt[self.groupAlts]

            # Return current xtuple when xid changes
            if self.lastXid is not None and (xid != self.lastXid or self.flatten):
                self.lastXid = xid
                break

            # Parse next result row
            self.lastXid = xid
            xtup.xid = self.lastXid
            xtup.append(self.__parseAlternative(self.lastAlt))

        # Do not adjust enc, alts,.. for 'view table'
        if not self.isViewTable:
            self.__checkVersions(xtup)

        # select merged
        if self.merged:
            xtup.alternatives = self.__horizontalMerging(xtup.alternatives)

        return xtup

    def __getLin(self, trio_aids):
        lin = xtuple.Lineage()
        if self.distinct:
            lin.append(self.distinctTable, int(trio_aids[0]))
        else:
            for i in range(len(trio_aids)):
                if isinstance(self.root_node.children[2].children[i], As):
                    lin.append(self.root_node.children[2].children[i].children[0].value, int(trio_aids[i]))
                else:
                    lin.append(self.root_node.children[2].children[i].value, int(trio_aids[i]))
        return lin

    def __getXid(self, alt):
        return "_".join("%s" % xid for xid in alt[self.num_datavalues:self.num_datavalues + self.num_joins])

    def __checkVersions(self, xtup):
        for a1 in xtup.alternatives:
            alts = 0
            questionMark = False
            for a2 in xtup.alternatives:
                    alts += 1
                    questionMark = questionMark or a2.alts < a2.enc
            questionMark = questionMark or alts < a1.alts
            a1.alts = alts
            if questionMark and not self.nomaybe:
                a1.enc = alts + 1
            else:
                a1.enc = alts
            if self.groupAlts > -1 or self.nomaybe:
                a1.confidence = None
            if self.gotUniform:
                a1.confidence = 1.0 / a1.enc

    def __parseAlternative(self, alt):
        for i in range(self.num_datavalues):
            if isinstance(alt[i], str):
                alt[i] = unicode(alt[i], 'utf-8')

        if self.nolineage:
            a = xtuple.Alternative(alt[:self.num_datavalues])
        else:
            a = xtuple.Alternative(alt[:self.num_datavalues],
                                   self.__getLin(alt[self.num_datavalues + self.num_joins
                                                     :self.num_datavalues + 2 * self.num_joins]))
        a.alts = int(alt[self.num_datavalues + 2 * self.num_joins])
        a.enc = int(alt[self.num_datavalues + 2 * self.num_joins + 1])
        if not self.noconf and not self.gotUniform \
           and alt[self.num_datavalues + 2 * self.num_joins + 2] is not None:
            a.confidence = float(alt[self.num_datavalues + 2 * self.num_joins + 2])
        return a

    def __horizontalMerging(self, alternatives):
        isUnique = []
        merged_alternatives = []
        for i in range(len(alternatives)):
            isUnique = isUnique + [0]
        for i in range(len(alternatives)):
            if isUnique[i] == 0:
                disj = 0
                j = i + 1
                while j < len(alternatives):
                    if (common.compare_values(alternatives[i].data, alternatives[j].data, self.num_datavalues) == 0
                       ):
                        # Track disjunctive lineage
                        disj += 1
                        for lin in alternatives[j].getLineage().items():
                            alternatives[i].lineage.append(lin[0], lin[1], disj)
                        if alternatives[i].getConfidence() is not None \
                           and alternatives[j].getConfidence() is not None:
                            alternatives[i].setConfidence(alternatives[i].getConfidence() + alternatives[j].getConfidence())
                        isUnique[i] += 1
                        isUnique[j] += 1
                    j += 1
                merged_alternatives.append(alternatives[i])

        # Normalize encoding and update xtuple alternatives
        for a in merged_alternatives:
            if a.enc > a.alts:
                a.enc = a.alts + 1

        return merged_alternatives
