__id__ = "$Id$"

class Lineage:
    """Represent a lineage of an alternative in Trio."""
    def __init__( self ):
        """Constructor."""
        self.srctable = []
        self.srcaid = []
        self.disjlin = [] #delta of 0 means conjunctive, 1 means disj, 2 means "and not"

    def append(self, table, aid, disjlin=0):
        self.srctable.append(table)
        self.srcaid.append(aid)
        self.disjlin.append(disjlin)

    def tables(self):
        return self.srctable

    def aids(self):
        return self.srcaid

    def disj(self):
        return self.disjlin

    def items(self):
        return map(None, self.srctable, self.srcaid, self.disjlin)

    def aidstrs(self):
        return [str(x) for x in self.srcaid]

class Alternative:
    """Represent an alternative in Trio.
    @ivar data: Attribute values for this alternative.
    @type data: A list.
    @ivar lineage: Lineage of this alternative.
    @type lineage: A L{Lineage} object.
    @ivar confidence: Confidence value for this alternative.
    @type confidence: float.
    """

    def __init__(self, data, lineage=Lineage(), alts=1, enc=1, conf=None):
        """Constructor for Alternative objects"""
        self.data = data
        self.lineage = lineage
        self.alts = alts
        self.enc = enc
        self.confidence = conf
        self.start = -1;
        self.end = -1;

    def __getitem__(self, index):
        """Return an attribute at the location specified by index."""
        return self.data[index]

    def __delitem__(self, index):
        """delete item"""
        del self.data[index]

    def __len__(self):
        """Return number of attributes in this alternative."""
        return len(self.data)

    def __str__(self):
        result = "(%s)" % ",".join(str(x) for x in self.data)
        if self.confidence is not None:
            result += ":%.3f" % self.confidence
        elif self.alts < self.enc:
            result += "?"
        return result

    def append(self, attr):
        """Append an attribute value to the alternative"""
        self.data.append(attr)

    def getLineage(self):
        return self.lineage

    def getConfidence(self):
        return self.confidence

    def traceLineage(self, tcnx):
        tnc = tcnx.ncursor()
        lin = self.lineage.items()
        # transient result
        if len(lin) > 1:
            return lin
        # persistent result
        elif len(lin) == 1:
            query = "SELECT srctable, srclin, disjlin FROM trio_lin_%s WHERE aid = %s ORDER BY disjlin asc" \
                    % (lin[0][0], lin[0][1])
            tnc.execute(query)
            return tnc.fetchall()
        else:
            return []

    def computeConfidence(self, tcnx):
        tnc = tcnx.ncursor()
        query = "SELECT conf(array[%s], array[%s], array[%s], NULL, 0, 1)" \
                % (", ".join("'%s'" % x for x in self.lineage.tables()),
                   ", ".join("%s" % str(x) for x in self.lineage.aids()),
                   ", ".join("%s" % str(x) for x in self.lineage.disj()))
        conf = None
        try:
            tnc.execute(query)
            conf = tnc.fetchone()[0]
        except Exception, ex:
            tcnx.rollback()
            conf = "error"
        return conf

    def setConfidence(self, conf):
        self.confidence = conf

    def getQuestionMark(self):
        return self.enc > self.alts

    def setQuestionMark(self):
        self.enc = self.alts + 1

class XTuple:
    def __init__(self):
        """A constructor for XTuple."""
        self.alternatives = []

    def __str__(self):
        alts = [str(x) for x in self.alternatives]
        alts.sort()
        return "|".join(alts) + "\n"

    def __getitem__(self, index):
        """Returns the alternative at the specified index."""
        return self.alternatives[index]

    def __len__(self):
        """Return number of alternatives in this object."""
        return len(self.alternatives)

    def getAlternative(self, index):
        """Returns the alternative at the specified index."""
        return self.alternatives[index]

    def append(self, alternative):
        """Append alternative to this xtuple.

        @param alternative: Alternative to be appended to this xtuple.
        @type alternative: Alternative
        """
        self.alternatives.append(alternative)

    def extend(self, alternatives):
        """Append alternatives to this xtuple.

        @param alternatives: Alternatives to be appended to this xtuple.
        @type alternatives: L{Alternative}
        """
        self.alternatives.extend(alternatives)

    def getConfidence(self, version=None):
        """Return sum of confidence values."""
        c = None
        for a in self.alternatives:
                if a.getConfidence() is None:
                    return None
                elif c is None:
                    c = 0
                c += a.getConfidence()
        return c

    def getQuestionMark(self, version=None):
        if self.getConfidence(version) is not None:
            return self.getConfidence(version) < 1.0
        else:
            for a in self.alternatives:
                    if a.getQuestionMark():
                        return True
            return False

    def setQuestionMark(self, version=None):
        for a in self.alternatives:
                a.setQuestionMark()
