#!/usr/bin/env python

__id__ = "$Id$"

from getpass import getpass
import optparse
import signal
import sys
import time
from trio import triodb
from trio import common

try:
    import readline
    import os
    import atexit
    histfile = os.path.join(os.environ["HOME"], ".trio_history")
except:
    histfile = None

trio_options = {"aids" : ["on", "off"]}

def lin_command(comm, trio_cursor):
    try:
        c = trio_cursor.tdb.pdb.cursor()
        comm = comm.split()
        if len(comm) < 3 or len(comm) > 4 or comm[1] != "lineage":
            raise Exception("Invalid explain command. Use: explain lineage <tablename> [aid]")
        if len(comm) == 3:
            query = "VIEW TABLE %s EXPLAIN LINEAGE" % (comm[2])
        else:
            query = "VIEW TABLE %s WHERE trio_aid = %s EXPLAIN LINEAGE" % (comm[2], comm[3])
        trio_cursor.execute(query)
        if trio_cursor.rowcount == 0:
            raise Exception("Invalid aid for table '%s'" % comm[2])
        print_result(trio_cursor)
        c.close()
    except Exception, e:
        print e

def conf_command(comm, trio_cursor):
    try:
        comm = comm.split()
        if len(comm) < 3 or len(comm) > 4 or comm[1] != "confidences":
            raise Exception("Invalid compute confidences command. Use: compute confidences <tablename> [aid]")
        if len(comm) == 3:
            query = "VIEW TABLE %s COMPUTE CONFIDENCES" % comm[2]
        else:
            query = "VIEW TABLE %s WHERE trio_aid = %s COMPUTE CONFIDENCES" % (comm[2], comm[3])
        trio_cursor.execute(query)
        if trio_cursor.rowcount == 0:
            raise Exception("Invalid aid for table '%s'" % comm[2])
        print_result(trio_cursor)
    except Exception, e:
        print e

def set_command(comm):
    try:
        (attr, value) = comm.split()[1:]
        if not (attr in trio_options and value in trio_options[attr]):
            raise Exception()
        common.cp.set("global", attr, value)
        print '"%s" set to "%s" ' % (attr, value)
    except:
        print("Invalid set command.")

def main_loop(cnx, args=[]):
    c = cnx.cursor()
    n = cnx.pdb.cursor()
    commands = ""
    while True:
        if len(args) == 0:
            try:
                line = raw_input("> ")
            except EOFError:
                break
            commands += " " + line
            parts = commands.split(";")
            commands = parts[-1:][0]
            comm_list = parts[:-1]
        else:
            comm_list = args
        for comm in comm_list:
            comm = comm.replace('\r', ' ')
            if comm.isspace():
                continue
            try:
                #print "\n" + str(comm) +
                c.start_time = time.time()
                if comm.strip().lower() == "commit":
                    cnx.commit()
                elif comm.strip().lower() == "rollback":
                    cnx.rollback()
                elif comm.strip().lower().startswith("set"):
                    set_command(comm.strip().lower())
                elif comm.strip().lower().startswith("compute"):
                    conf_command(comm.strip().lower(), c)
                elif comm.strip().lower().startswith("explain"):
                    lin_command(comm.strip().lower(), c)
                else:
                    c.execute(comm)
                    if c.isFetchable:
                        print_result(c)

                c.end_time = time.time()
                if len(args) == 0:
                    print "\nTOTAL:  " + str(c.end_time - c.start_time) + "\n"
            except Exception, e:
                common.print_exception()
                continue
        if len(args) > 0:
            break

def print_result(cursor):
    attrnames = []
    colwidth = 21
    if cursor.isXCursor and common.cp.get("global", "aids") == "on" and cursor.isViewTable:
        cursor.description = [["aid"]] + cursor.description
    for i in range(len(cursor.description)):
        if cursor.u_columns is not None and i in cursor.u_columns:
            attrnames.append(cursor.description[i][0].upper() + "*")
        else:
            attrnames.append(cursor.description[i][0].upper())
    numdashes = (colwidth + 2) * len(cursor.description)
    sys.stdout.write("\n+")
    for i in range(numdashes):
        sys.stdout.write( "-" )
    sys.stdout.write("+\n")
    sys.stdout.write( "| " )
    for i in range(len(attrnames)):
        formatstring = "%%%ds |" % colwidth
        sys.stdout.write(formatstring % attrnames[i])
    sys.stdout.write("\n+%s+\n" % ("=" * numdashes))
    if cursor.isXCursor:
        print_xtuples(cursor, colwidth)
    else:
        print_tuples(cursor, colwidth)

def print_tuples(cursor, colwidth):
    formatstring = "%%%ds |" % colwidth
    numdashes = (colwidth + 2) * len(cursor.description)
    while True:
        tuple = cursor.fetchone()
        if tuple is None:
            break
        for i in range(len(tuple)):
            sys.stdout.write( "| " )
            sys.stdout.write(formatstring % str(tuple[i])[:colwidth])
            print
            sys.stdout.write("+%s+\n" % ("-" * numdashes))

def print_xtuples(cursor, colwidth):
    numdashes = (colwidth + 2) * len(cursor.description)
    while True:
        xtuple = cursor.fetchone()
        if xtuple is None:
            break
        for i in range(len(xtuple)):
            if common.cp.get("global", "aids") == "on" and cursor.isViewTable:
                xtuple[i].data = [xtuple[i].lineage.srcaid[0]] + xtuple[i].data
            sys.stdout.write( "| " )
            j = 0
            for a in xtuple[i]:
                formatstring = "%%%ds |" % colwidth
                if isinstance(a, unicode):
                    sys.stdout.write(formatstring % a.encode('utf-8')[:colwidth])
                elif a is not None:
                    sys.stdout.write(formatstring % str(a)[:colwidth])
                else:
                    sys.stdout.write(formatstring % "NULL")
                j += 1
            conf = xtuple[i].getConfidence()
            quest = xtuple[i].getQuestionMark()
            if conf is not None and conf >= 0:
                sys.stdout.write(str(conf))
            elif quest:
                sys.stdout.write("?")
            else:
                sys.stdout.write(" ")
            if cursor.hasExplainLineage:
                sys.stdout.write(" " + print_lineage(cursor, xtuple[i].traceLineage(cursor.tdb)))
            print
        sys.stdout.write( "+%s+\n" % ( "-" * numdashes ) )

def print_lineage(cursor, lineage):
    tnc = cursor.tdb.ncursor()
    s = ""
    for i in range(len(lineage)):
        query = "SELECT srctable, srclin, disjlin FROM trio_lin_%s WHERE aid = %s ORDER BY disjlin asc" \
                % (lineage[i][0], lineage[i][1])
        tnc.execute(query)
        nested_lineage = tnc.fetchall()
        if i == 0 and lineage[-1][2] != lineage[0][2]:
            s+= "["
        if len(nested_lineage) > 0:
            s += "[" + print_lineage(cursor, nested_lineage) + "]"
        else:
            s += lineage[i][0] + ":" + str(lineage[i][1])
        if i == len(lineage) -1 and lineage[-1][2] != lineage[0][2]:
            s += "]"
        if len(lineage) > 1 and i < len(lineage) - 1 and lineage[i+1][2] - lineage[i][2] == 0:
            s += " AND "
        elif len(lineage) > 1 and i < len(lineage) - 1 and lineage[i+1][2] - lineage[i][2] == 1:
            s += "] OR ["
        elif len(lineage) > 1 and i < len(lineage) - 1 and lineage[i+1][2] - lineage[i][2] == 2:
            s += "] AND NOT ["
        elif i < len(lineage) - 1:
            s += " AND "
    return s

def main(argv=None):
    if argv is None:
        argv = sys.argv

    p = optparse.OptionParser()
    p.add_option("-d", dest="database", help="database name to connect to")
    p.add_option("-u", dest="user", help="database user name to connect as")
    p.add_option("-p", dest="password", action="store_true", help="prompt for password")

    (options, args) = p.parse_args()

    options = options.__dict__
    if options["password"] == True:
        try:
            options["password"] = getpass('Enter password: ')
        except EOFError:
            return 0

    signal.signal(signal.SIGINT, signal.SIG_IGN) # ignores ctl-c

    if histfile is not None:
        try:
            readline.read_history_file(histfile)
        except IOError:
            pass
        atexit.register(readline.write_history_file, histfile)

    try:
        cnx = triodb.connect(**options)
        if len(args) == 0:
            print
            print "Welcome to the TrioPlus command-line terminal. Press <CTRL><D> to exit."
            print
        main_loop(cnx, args)
        cnx.close()
    except:
        common.print_exception()
        return 1

if __name__ == "__main__":
    sys.exit(main())
